import numpy as np
from abc import ABCMeta, abstractmethod

# Initial condition
class Advection(object):
    __metaclass__ = ABCMeta

    def __init__(self,a,tf,dt):
        self.a = a
        self.tf = tf
        self.dt = dt

    @abstractmethod
    def init_advection(self,s):
        return

    @abstractmethod
    def exact_advection(self,s,t):
        return

    @abstractmethod
    def solver(self,s,*args,**kwargs):
        return

class AdvectionSimple(Advection):

    def __init__(self,a,tf,dt):
        Advection.__init__(self,a,tf,dt)

# Initial condition (sine function)
    def init_advection(self,s):
        g = np.mgrid[[slice(0,1,1./(2**s[i]+1)) for i in range(2)]]
        buf=1.0
        for d in range(2):
            buf*=np.sin(2*np.pi*g[d])#*g[0]#+r
        return buf+np.pi

# Exact solution (shifted sine function)
    def exact_advection(self,s,t):
        g = np.mgrid[[slice(0,1,1./(2**s[i]+1)) for i in range(2)]]
        buf=1.0
        for d in range(2):
            buf*=np.sin(2*np.pi*(g[d]-self.a[d]*t))#*g[0]#+r
        return buf+np.pi

# Solver: Lax Wendroff 2D
    def solver(self,s,f=None,Nt=0,sdcFactor=1, sdcIter=-1,sdcSub='highest'):

        if f == None:
            f = self.init_advection(s)
        
        dX = 1./(2**np.array(s)+1)
        
        cu = self.dt*self.a[0]/dX[0]
        cv = self.dt*self.a[1]/dX[1]

        r = lambda i,d: np.roll(f,i,axis=d)
        if Nt == 0:
            Nt = int(np.ceil(self.tf/self.dt))

        if sdcIter >= Nt:
           raise AssertionError('The SDC iteration cannot be larger than Nt.') 

	# Split the iterations from nt = [0,sdcIter] and nt = [sdcIter+1,Nt],
	# and inject SDC in between
        for nt in range(sdcIter+1):
            fiplusm = 0.5*(f+r(-1,0)) - 0.5*cu*(r(-1,0)-f)
            fjplusm = 0.5*(f+r(-1,1)) - 0.5*cv*(r(-1,1)-f)
            fiminusm = 0.5*(r(1,0)+f) - 0.5*cu*(f-r(1,0))
            fjminusm = 0.5*(r(1,1)+f) - 0.5*cv*(f-r(1,1))
            f = f -cu*(fiplusm - fiminusm) - cv*(fjplusm - fjminusm) + cu*cv*(f-r(1,1)-r(1,0)+np.roll(np.roll(f,1,0),1,1))

        # Inject SDC
	# Worst case scenario: affected point is in the highest hierarchical subspace
        if sdcSub == 'highest':
           # [i1,i2] = 2**np.array(s)-1
            [i1,i2] = 2**np.array(s)/2-1
	# Point in the middle of the domain
        if sdcSub == 'lowest':
            [i1,i2] = 2**np.array(s)/2

			  
        if sdcFactor != 1:
            # Multiply the value by a factor
            f[i1,i2] *= sdcFactor
            # Set the value to the factor
           # f[i1,i2] = sdcFactor

        for nt in range(sdcIter+1,Nt):
            fiplusm = 0.5*(f+r(-1,0)) - 0.5*cu*(r(-1,0)-f)
            fjplusm = 0.5*(f+r(-1,1)) - 0.5*cv*(r(-1,1)-f)
            fiminusm = 0.5*(r(1,0)+f) - 0.5*cu*(f-r(1,0))
            fjminusm = 0.5*(r(1,1)+f) - 0.5*cv*(f-r(1,1))
            f = f -cu*(fiplusm - fiminusm) - cv*(fjplusm - fjminusm) + cu*cv*(f-r(1,1)-r(1,0)+np.roll(np.roll(f,1,0),1,1))

        return f
