import numpy as np
import itertools as it
import ErrorNorms as eN

def computeBound2D(s,t,p,kappa=1):
    hs = 2.0**(-np.array(s)*p)
    ht = 2.0**(-np.array(t)*p)
    key_common = tuple(map(min,zip(*[s,t])))
    factor = 4
    return factor*kappa*(sum(hs)+sum(ht) + np.prod(hs) + np.prod(ht))

def computeTightBound2D(s,t,p,kappa=1):
    hs = 2.0**(-np.array(s)*p)
    ht = 2.0**(-np.array(t)*p)
    key_common = tuple(map(min,zip(*[s,t])))
    factor = 4
    return factor*kappa*(abs(ht[0] - hs[0])+abs(ht[1] - hs[1]))

def getMinDt(keys,a):
    dt = 1000.
    factor = .5
    for s in keys:
        dX = 1./(2.0**np.array(s)+1)
        dt_tmp = min(factor*dX[0]/a[0],factor*dX[1]/a[1])
        if dt_tmp < dt:
            dt = dt_tmp
    return dt

def sortActiveSet(activeSet,dim):
    sortedKeys = sorted(activeSet, key=itemgetter(*range(dim)),reverse=True)
    return sortedKeys

def calculateKappa(grid_s,grid_t):
    key_s = grid_s.getLevelvector()
    key_t = grid_t.getLevelvector()
    key_common = tuple(map(min,zip(*[key_s,key_t])))
    bound = computeBound2D(key_s,key_t,2)
    kappa = np.abs(grid_s.getNodalSubspace(key_common,'h')-grid_t.getNodalSubspace(key_common,'h'))
    kappa = kappa/bound
    return kappa
    
def calculateBeta(grid_s,grid_t,withExtrapolation=False,lnorm=np.inf):
    key_s = grid_s.getLevelvector()
    key_t = grid_t.getLevelvector()
    if withExtrapolation:
        key_common = tuple(map(max,zip(*[key_s,key_t])))
    else:
        key_common = tuple(map(min,zip(*[key_s,key_t])))
    list_ranges = [range(1,key_common[0]+1),range(1,key_common[1]+1)]
    betas_all = {}
    for subspace in it.product(*list_ranges):
        betas_all[subspace] = calculateBetaL(grid_s,grid_t,subspace,lnorm=lnorm)

    maxSubs = max(betas_all, key=betas_all.get)
    return betas_all[maxSubs]

def calculateBetaL(grid_s,grid_t,level,lnorm=np.inf):
    betaL = np.linalg.norm(grid_s.getHierSubspace(level,'h').flatten()-grid_t.getHierSubspace(level,'h').flatten(),lnorm)
    norm_s = np.linalg.norm(grid_s.getHierSubspace(level,'h').flatten(),lnorm)
    norm_t = np.linalg.norm(grid_t.getHierSubspace(level,'h').flatten(),lnorm)
    factor = min(norm_s,norm_t)
    if factor < 1e-14: # Happens when extrapolating and then hierarchizing
        factor = 1

    return betaL/factor

def calculateDiff(grid_s,grid_t,lmax):
    key_s = grid_s.getLevelvector()
    key_t = grid_t.getLevelvector()
    key_max = (max(key_s[0],key_t[0]),max(key_s[1],key_t[1]))
    diff = eN.L2RelativeError(grid_s.extrapolateData(key_max),grid_t.extrapolateData(key_max))
    return diff

def calculateD(l,s,t):
    return 4*pow(2.,2*(max(l)))*computeBound2D(s,t,2)

def computeDifference(s,t,p,a=2.):
    hs = 2.0**(-np.array(s)*p)
    ht = 2.0**(-np.array(t)*p)
    D1s = pow(2.,s[0]*p-a*sum(s)-2)
    D2s = pow(2.,s[1]*p-a*sum(s)-2)
    D1t = pow(2.,t[0]*p-a*sum(t)-2)
    D2t = pow(2.,t[1]*p-a*sum(t)-2)
    return (abs(ht[0] + hs[0]) + abs(ht[1] + hs[1]))*max([D1s,D2s,D1t,D2t])

def boundVariance(sub,scheme):
    largerSubs = scheme.getLargerSubspaces(sub,unified=False)
    g_sum = 0
    Nl = len(largerSubs)
    for s,t in it.combinations(largerSubs,2):
        g_sum += computeBound2D(s,t,2)**2/16.
#     for s in largerSubs:
#         for t in largerSubs:
#             if not all(s == t):
#                 g_sum += computeBound2D(s,t,2)**2/16.
    
    return 16*g_sum/Nl**2

def is_outlier(points, thresh=3.5):
    """
    Returns a boolean array with True if points are outliers and False 
    otherwise.

    Parameters:
    -----------
        points : An numobservations by numdimensions np.array of observations
        thresh : The modified z-score to use as a threshold. Observations with
            a modified z-score (based on the median absolute deviation) greater
            than this value will be classified as outliers.

    Returns:
    --------
        mask : A numobservations-length boolean array.

    References:
    ----------
        Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor. 
    """
    if len(points.shape) == 1:
        points = points[:,None]
    median = np.median(points, axis=0)
    diff = np.sum((points - median)**2, axis=-1)
    diff = np.sqrt(diff)
    med_abs_deviation = np.median(diff)

    modified_z_score = 0.6745 * diff / med_abs_deviation

    return modified_z_score > thresh

def Grubbs_outlier_test(y_i, alpha=0.95):
    """
    Perform Grubbs' outlier test.
    
    ARGUMENTS
    y_i (list or numpy array) - dataset
    alpha (float) - significance cutoff for test

    RETURNS
    G_i (list) - Grubbs G statistic for each member of the dataset
    Gtest (float) - rejection cutoff; hypothesis that no outliers exist if G_i.max() > Gtest
    no_outliers (bool) - boolean indicating whether there are no outliers at specified significance level
    index (int) - integer index of outlier with maximum G_i    
    """
    s = y_i.std()
    G_i = np.abs(y_i - y_i.mean()) / s
    N = y_i.size
    from scipy import stats
    t = stats.t.isf(1 - alpha/(2*N), N-2)
    Gtest = (N-1)/np.sqrt(N) * np.sqrt(t**2 / (N-2+t**2))    
    G = G_i.max()
    index = G_i.argmax()
    no_outliers = (G > Gtest)
    return [G_i, Gtest, no_outliers, index]

# Returns the hierarchical subspaces of level s of a given array A
def getHierSubs(A,s):
    old_s = np.log2(np.array(A.shape)-1)
    ind = []
    for i in range(2):
        if s[i] == 1:
            factor = int(2**(old_s[i]-s[i]))
            ind.append(slice(None,None,factor))
        else:
            factor = int(2**(old_s[i]-s[i]+1))
            ind.append(slice(factor>>1,None,factor))
    return A[tuple(ind)]

# Returns the grid points of a full grid of level "level"
def fun(level):
    s = 2**np.array(level)+1
    meshgrid = np.mgrid[0:1:1j*s[0],0:1:1j*s[1]]
    return meshgrid

# Helper function to plot an arbitrary combination technique
def plot_combination_technique( lmin, lmax, combidict, combigrids, fig, plusMinus=True, withFaults=False, faults=[], s=5, lw=.01, cl='blue'):
    from itertools import product
    import matplotlib.pyplot as plt
    lm = np.array(lmax)
    ll = lmax[0] + lmin[0]
    list_ranges = [range(1,lmax[0]+1),range(1,lmax[1]+1)]
    for key in product(*list_ranges):
        k = np.array(key)
        ax = plt.subplot2grid(lm,(lm[1]-k[1],k[0]-1))
        grid = combigrids.gridsDict[key,]
        data = grid.getData()
        if key in combidict.keys():
            pmesh = plt.pcolormesh(data.transpose())
            if (plusMinus):
                if sum(key) == ll:
                    plt.plot(data.shape[0]/2.,data.shape[1]/2.,marker='+',ms=70,mew=16,color='k',alpha=.8)
                else:
                    plt.plot(data.shape[0]/2.,data.shape[1]/2.,marker='_',ms=70,mew=16,color='k',alpha=.8)
            pmesh.axes.set_xlim(0, data.shape[0])
            pmesh.axes.set_ylim(0, data.shape[1])
        else:
            plt.scatter(data[0].flatten(), data[1].flatten(),s=s,linewidths=lw,c=cl)
        if withFaults and key in faults:
            plt.plot(data.shape[0]/2.,data.shape[1]/2.,marker='x',ms=72,mew=20,color='r')
        ax.set_xticks([])
        ax.set_yticks([])

    fig.text(0.5, 0.05, '$l_1$', fontsize = 24)
    fig.text(0.05, 0.5, '$l_2$', fontsize = 24)

# Helper function to plot solutions (exact, full, and combined)
def plot_all_solutions( fig, solution_exact, solution_combi, solution_full=[]):
    fig.subplots_adjust(hspace=0.3)

    def plot_one_solution( solution, position, title_text ):
        import matplotlib.pyplot as plt
        if solution_full == []:
            plt.subplot2grid( (1,2), position )
        else:
            plt.subplot2grid( (1,3), position )
        plt.title( title_text, fontsize=26 )
        myMesh = plt.pcolormesh( solution, rasterized=True )
        myMesh.axes.set_xlim( 0, solution.shape[0] )
        myMesh.axes.set_ylim( 0, solution.shape[1] )
        myMesh.axes.tick_params(axis='both', which='major', labelsize=12)
        plt.colorbar()

    # Plot exact solution
    plot_one_solution( solution_exact, (0,0), 'Exact solution' )
    
    # Plot combined solution
    plot_one_solution( solution_combi, (0,1), 'Combined solution' )
   
    if solution_full != []: 
        plot_one_solution( solution_full, (0,2), 'Full solution' )

def parabola_2d(l):
    g = np.mgrid[[slice(0,1+1./(2**l[i]+1),1./(2**l[i])) for i in range(2)]]
    buf = 4*g[0]*(g[0]-1)
    buf *= 4*g[1]*(g[1]-1)
    return buf
